# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from dataclasses import dataclass
from dataclasses import field
from enum import Enum

import numpy

from orso.exceptions import DataValidationError
from orso.tools import random_string
from orso.types import ORSO_TO_PYTHON_MAP
from orso.types import OrsoTypes


@dataclass
class FlatColumn:
    # This is a standard column type.
    # Unlike the other column types, we don't store the values for this Column
    # here, we read them from the underlying data structure (orso/pyarrow/velox).

    name: str
    type: OrsoTypes
    description: typing.Optional[str] = None
    aliases: typing.Optional[typing.List[str]] = field(default_factory=list)
    nullable: bool = True
    expectations: typing.Optional[list] = field(default_factory=list)
    identity: str = field(default_factory=random_string)

    def __str__(self):
        return self.identity


@dataclass
class ConstantColumn(FlatColumn):
    """
    Rather than pass around columns of constant values, where we can we should
    replace them with this column type.

    note we don't implement anything here which deals with doing operations on
    two constant columns; whilst that would be a good optimization, the better
    way to do this is in the query optimizer, do operations on two constants
    while we're still working with a query plan.
    """

    length: int = 0
    value: typing.Any = None

    def materialize(self):
        """
        Turn this virtual column into a list
        """
        return numpy.array([self.value] * self.length)


@dataclass
class DictionaryColumn(FlatColumn):
    """
    If we know a column has a small amount of unique values AND is a large column
    AND we're going to perform an operation on the values, we should dictionary
    encode the column. This allows us to operate once on each unique value in the
    column, rather than each value in the column individually. At the cost of
    constructing and materializing the dictionary encoding.
    """

    values: typing.List[typing.Any] = field(default_factory=list)

    def __post_init__(self):
        """
        Perform the encoding for this column
        """
        values = numpy.asarray(self.values)
        self.dictionary, self.encoding = numpy.unique(values, return_inverse=True)

    def materialize(self):
        """
        Turn this virtual column into a list
        """
        return self.dictionary[self.encoding]


@dataclass
class RelationSchema:
    name: str
    aliases: typing.List[str] = field(default_factory=list)
    columns: typing.List[FlatColumn] = field(default_factory=list)

    @property
    def num_columns(self):
        return len(self.columns)

    def find_column(self, column_name: str):
        """find a column by name or alias"""
        for column in self.columns:
            if column.name == column_name:
                return column
            if column_name in column.aliases:
                return column
        return None

    def all_column_names(self):
        """return all the names for columns in this relation"""

        def _inner():
            for column in self.columns:
                yield column.name
                yield from column.aliases

        return list(_inner())

    def column(self, i):
        """get column by name or index"""
        if isinstance(i, int):
            return self.columns[i]
        else:
            return self.find_column(i)

    def to_dict(self):
        """Convert a Schema to a dictionary"""
        from dataclasses import asdict

        def _converter(obj):
            """handle enum serialization"""
            return {key: value.value if isinstance(value, Enum) else value for key, value in obj}

        return asdict(self, dict_factory=_converter)

    @classmethod
    def from_dict(cls, dic):
        """Create a Schema from a dictionary"""
        schema = RelationSchema(
            name=dic["name"],
            aliases=dic.get("aliases", []),
        )
        for column in dic["columns"]:
            schema.columns.append(FlatColumn(**column))
        return schema

    def validate(self, data: dict):
        """
        Perform schema validation against a dictionary formatted record
        """
        # If it's not dictionary-like, it's not valid
        if not isinstance(data, typing.MutableMapping):
            raise TypeError("Cannot validate non Dictionary-type value")

        for column in self.columns:
            # If the column is missing from the data, it's not valid
            if column.name not in data:
                raise DataValidationError(column=column, value=None, error="Column Missing")

            # Get the value for this field out of the dict
            value = data[column.name]

            # If the value is null, it's value if it's nullable
            if value is None:
                if not column.nullable:
                    raise DataValidationError(
                        column=column, value=value, error="None not acceptable"
                    )
            else:
                # finally, is it the right type
                if not isinstance(value, ORSO_TO_PYTHON_MAP.get(column.type)):
                    raise DataValidationError(column=column, value=value, error="Incorrect Type")

        return True



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from tests import cities


def test_can_profile():
    # fmt:off
    
    import orso
    df = orso.DataFrame(cities.values)
    profile = df.profile

    assert profile.shape == (6, 10), profile.shape
    assert profile.collect("count") == [20] * 6

    # we've seen problems that this has uncovered
    for i in range(5):
        profile = profile.profile


def test_opteryx_profile():
    try:
        sys.path.insert(1, os.path.join(sys.path[0], "../../opteryx"))
        import opteryx

        planets = opteryx.query("SELECT * FROM $planets")
        profile = planets.profile
        assert profile.shape == (20, 10), profile.shape
        assert profile.collect("count") == [9] * 20
    except ImportError:
        # if Opteryx isn't installed, don't fail
        pass


if __name__ == "__main__":  # pragma: no cover
    test_can_profile()
    test_opteryx_profile()

    print("✅ okay")



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import typing

from orso.exceptions import MissingDependencyError
from orso.row import Row


def to_arrow(dataset, size=None):
    try:
        import pyarrow
    except ImportError as import_error:
        raise MissingDependencyError(import_error.name) from import_error

    # Create a list of PyArrow arrays from the rows
    rows = dataset._rows
    arrays = [
        pyarrow.array(col) for col in zip(*(rows if size is None else itertools.islice(rows, size)))
    ]

    # Create a PyArrow table from the arrays and schema
    if arrays:
        table = pyarrow.Table.from_arrays(arrays, dataset.column_names)
    else:
        table = pyarrow.Table.from_arrays([[]] * len(dataset.column_names), dataset.column_names)

    return table


def from_arrow(tables, size=None):
    """
    Convert an Arrow table or an iterable of Arrow tables to a generator of
    Python objects.
    """
    from orso.tools import parquet_type_map

    if not isinstance(tables, (typing.Generator, list, tuple)):
        tables = [tables]

    if isinstance(tables, (list, tuple)):
        tables = iter(tables)

    # Extract schema information from the first table
    first_table = next(tables, None)
    if first_table is None:
        return [], {}

    schema = first_table.schema
    fields = {
        str(field.name): {"type": parquet_type_map(field.type), "nullable": field.nullable}
        for field in schema
    }

    # Create a generator of tuples from the columns
    row_factory = Row.create_class(fields)

    BATCH_SIZE: int = 10000
    if size:
        BATCH_SIZE = min(size, BATCH_SIZE)

    rows = []
    for table in [first_table] + list(tables):
        batches = table.to_batches(max_chunksize=BATCH_SIZE)
        for batch in batches:
            column_data_dict = batch.to_pydict()
            column_data = [column_data_dict[name] for name in schema.names]
            new_rows: typing.Iterable = [tuple()] * batch.num_rows
            for i, row_data in enumerate(zip(*column_data)):
                new_rows[i] = row_factory(row_data)
            rows.extend(new_rows)
            if size and len(rows) >= size:
                break

    # Limit the number of rows to 'size'
    if size:
        rows = itertools.islice(rows, size)

    return rows, fields


def to_pandas(dataset, size=None):
    try:
        import pandas
    except ImportError as import_error:
        raise MissingDependencyError(import_error.name) from import_error
    return pandas.DataFrame(r.as_dict for r in dataset.slice(0, size))


def from_pandas(pandas):
    raise NotImplementedError()


def to_polars(dataset, size=None):
    try:
        import polars
    except ImportError as import_error:
        raise MissingDependencyError(import_error.name) from import_error
    return polars.DataFrame(r.as_dict for r in dataset.slice(0, size))


def from_polars(polars):
    raise NotImplementedError()


def to_csv(dataset):
    raise NotImplementedError()


def from_csv(values):
    raise NotImplementedError()


def to_parquet(dataset):
    raise NotImplementedError()


def from_parquet(values):
    raise NotImplementedError()


def to_jsonl(dataset):
    raise NotImplementedError()


def from_jsonl(values):
    raise NotImplementedError()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import typing

from orso.schema import RelationSchema


class DataFrame:
    __slots__ = ("_schema", "_rows", "_cursor", "arraysize", "_row_factory")

    def __init__(
        self,
        dictionaries: typing.Optional[typing.List[dict]] = None,
        *,
        rows: typing.Union[typing.Generator, list, None] = None,
        schema: typing.Optional[RelationSchema] = None,
    ):
        from orso import Row

        if dictionaries is not None:
            if schema is not None or rows is not None:
                raise ValueError(
                    "Can't implicitly and explicitly define a DataFrame at the same time"
                )

            from itertools import chain

            # make the list of dicts iterable
            dicts = iter(dictionaries)

            # extract the first of the list, and get the types from it
            first_dict = {}
            first_dict = next(dicts)
            self._schema = {name: {"type": type(value)} for name, value in first_dict.items()}

            self._row_factory = Row.create_class(self._schema)
            # create a list of tuples
            self._rows: list = (  # type:ignore
                self._row_factory([row.get(k) for k in first_dict.keys()])
                for row in chain([first_dict], dicts)
            )
        else:
            self._schema = schema
            self._rows = rows  # type:ignore
            self._row_factory = Row.create_class(self._schema)
        self.arraysize = 100
        self._cursor = iter(self._rows or [])

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    @classmethod
    def from_arrow(cls, tables):
        from orso.converters import from_arrow

        rows, schema = from_arrow(tables)
        return cls(rows=rows, schema=schema)

    def arrow(self, size=None):
        from orso.converters import to_arrow

        return to_arrow(self, size=size)

    def pandas(self, size=None):
        from orso.converters import to_pandas

        return to_pandas(self, size)

    def polars(self, size=None):
        from orso.converters import to_polars

        return to_polars(self, size)

    def nbytes(self):
        self.materialize()
        for row in self._rows:
            size += len(row.to_bytes())
        return size

    def append(self, entry):
        self._schema.validate(entry)
        self._rows.append(self._row_factory(entry))
        self._cursor = None

    def head(self, size: int = 5):
        return self.slice(0, size)

    def tail(self, size: int = 5):
        return self.slice(offset=0 - size, length=size)

    def query(self, predicate):
        """
        Apply a Selection operation to a Relation, this filters the data in the
        Relation to just the entries which match the predicate.

        Parameters:
            predicate (callable):
                A function which can be applied to a tuple to determine if it
                should be returned in the target Relation.

        Returns:

        """
        # selection invalidates what we thought we knew about counts etc
        new_schema = {k: {"type": v.get("type")} for k, v in self._schema.items()}
        return DataFrame(rows=(r for r in filter(predicate, self._rows)), schema=new_schema)

    def select(self, attributes):
        if not isinstance(attributes, (list, tuple)):
            attributes = [attributes]
        attribute_indices = []
        new_header = {k: v for k, v in self._schema.items() if k in attributes}
        for index, attribute in enumerate(self._schema.keys()):
            if attribute in attributes:
                attribute_indices.append(index)

        def _inner_projection():
            for tup in self._rows:
                yield tuple([tup[indice] for indice in attribute_indices])

        return DataFrame(rows=_inner_projection(), schema=new_header)

    def materialize(self):
        if not isinstance(self._rows, list):
            self._rows = list(self._rows or [])

    def distinct(self):
        hash_list = {}

        def do_dedupe(data):
            for item in data:
                hashed_item = hash(item)
                if hashed_item not in hash_list:
                    yield item
                    hash_list[hashed_item] = True

        return DataFrame(rows=do_dedupe(self._rows), schema=self._schema)

    def collect(self, columns):
        single = False
        if not isinstance(columns, typing.Iterable) or isinstance(columns, str):
            single = True
            columns = [columns]

        # get the index of a column name
        columns = [c if isinstance(c, int) else self.column_names.index(c) for c in columns]

        # Initialize empty lists for each column
        result: tuple = tuple([] for _ in columns)

        # Extract the specified columns into the result lists
        for row in self._rows:
            for j, column in enumerate(columns):
                result[j].append(row[column])

        if single:
            return result[0]
        return result

    def slice(self, offset: int = 0, length: int = None):
        self.materialize()
        if offset < 0:
            offset = len(self._rows) + offset
        if length is None:
            return DataFrame(schema=self._schema, rows=self._rows[offset:])
        return DataFrame(schema=self._schema, rows=self._rows[offset : offset + length])

    def filter(self, mask):
        """
        Select rows from the DataFRame. The DataFrame is filtered based on boolean array.
        """
        return DataFrame(schema=self._schema, rows=(t for t, m in zip(self._rows, mask) if m))

    def take(self, indexes):
        """
        Select rows from the DataFrame. Rows are selected based on their appearance in the indexes
        list
        """
        return DataFrame(
            schema=self._schema, rows=(m for i, m in enumerate(self._rows) if i in indexes)
        )

    def row(self, i):
        self.materialize()
        return self._rows[i]

    def fetchone(self):
        if self._cursor is None:
            raise Exception("Cannot use fetchone and append on the same DataFrame")
        try:
            return next(self._cursor)
        except StopIteration:
            return None

    def fetchmany(self, size=None):
        if self._cursor is None:
            raise Exception("Cannot use fetchmany and append on the same DataFrame")
        fetch_size = self.arraysize if size is None else size
        entries = []
        for i in range(fetch_size):
            try:
                entry = next(self._cursor)
                entries.append(entry)
            except StopIteration:
                break
        return entries

    def fetchall(self):
        if self._cursor is None:
            raise Exception("Cannot use fetchall and append on the same DataFrame")
        return list(self._cursor)

    def display(
        self,
        limit: int = 5,
        display_width: typing.Union[bool, int] = True,
        max_column_width: int = 32,
        colorize: bool = True,
        show_types: bool = True,
    ):
        from .display import ascii_table

        return ascii_table(
            self,
            limit=limit,
            display_width=display_width,
            max_column_width=max_column_width,
            colorize=colorize,
            show_types=show_types,
        )

    def markdown(self, limit: int = 5, max_column_width: int = 30):
        from .display import markdown

        return "\n".join(markdown(self, limit=limit, max_column_width=max_column_width))

    def to_batches(self, batch_size: int = 1000):
        """
        Batch a DataFrame into batches of `size` records.

        Args:
            size (int): The size of each batch.

        Yields:
            DataFrames
        """
        self.materialize()
        for i in range(0, self.rowcount, batch_size):
            yield DataFrame(rows=self._rows[i : i + batch_size], schema=self._schema)

    @property
    def profile(self):
        from orso.profiler import DataProfile

        return DataProfile.from_dataset(self)

    @property
    def description(self):
        """
        name
        type_code
        display_size
        internal_size
        precision
        scale
        null_ok
        """
        from orso.types import PYTHON_TO_ORSO_MAP

        result = []
        for column in self.column_names:
            column_data = self._schema.get(column, {})
            data_type = PYTHON_TO_ORSO_MAP.get(column_data.get("type"))
            result.append(
                (
                    column,
                    None if data_type is None else data_type.name,
                    None,
                    None,
                    None,
                    None,
                    column_data.get("nullable"),
                )
            )
        return result

    @property
    def column_names(self):
        return tuple(self._schema.keys())

    @property
    def columncount(self):
        return len(self._schema.keys())

    @property
    def shape(self):
        return (self.rowcount, self.columncount)

    @property
    def rowcount(self):
        self.materialize()
        return len(self._rows)

    def __hash__(self):
        from orso.cityhash import CityHash32

        _hash = 0
        for row in self._rows:
            row_hash = CityHash32(str(row).encode())
            _hash = _hash ^ row_hash
        return _hash

    def __iter__(self):
        return iter(self._rows)

    def __len__(self) -> int:
        self.materialize()
        return len(self._rows)

    def __str__(self) -> str:
        from .display import ascii_table

        return (
            ascii_table(self, show_types=True)
            + f"\n[ {self.rowcount} rows x {self.columncount} columns ]"
        )

    def __repr__(self) -> str:
        size: int = 10
        try:
            from IPython import get_ipython

            i_am_in_a_notebook = get_ipython() is not None
        except ImportError:
            i_am_in_a_notebook = False

        if i_am_in_a_notebook:
            from IPython.display import HTML
            from IPython.display import display

            from .display import html_table

            html = html_table([r.as_dict for i, r in enumerate(iter(self)) if i < size], size)
            display(HTML(html))
            return ""
        else:
            from .display import ascii_table

            return (
                ascii_table(self, limit=size, top_and_tail=True)
                + f"\n[ {self.rowcount} rows x {self.columncount} columns ]"
            )

    def __add__(self, the_other):
        if self._schema != the_other._schema:
            raise ValueError("Schemas must be identical to add DataFrames")
        return DataFrame(rows=self._rows + the_other._rows, schema=self._schema)



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.dataframe import DataFrame
from tests import cities


def test_to_pandas():
    import pandas

    odf = DataFrame(cities.values)
    pdf = odf.pandas()

    assert len(pdf) == 20
    assert "name" in pdf.columns
    assert type(pdf) == pandas.DataFrame

    pdf = odf.pandas(size=4)
    assert len(pdf) == 4
    assert "name" in pdf.columns
    assert type(pdf) == pandas.DataFrame


if __name__ == "__main__":  # pragma: no cover
    test_to_pandas()
    print("✅ okay")



import os
import sys

import pyarrow

sys.path.insert(1, os.path.join(sys.path[0], ".."))
sys.path.insert(1, os.path.join(sys.path[0], "../.."))
sys.path.insert(1, os.path.join(sys.path[0], "../../opteryx"))

from orso import converters
from orso.dataframe import DataFrame


def test_from_arrow():
    # Create some sample data
    table1 = pyarrow.Table.from_arrays(
        [
            pyarrow.array([1, 2, 3]),
            pyarrow.array(["a", "b", "c"]),
            pyarrow.array([True, False, True]),
        ],
        ["id", "name", "active"],
    )
    table2 = pyarrow.Table.from_arrays(
        [
            pyarrow.array([4, 5, 6]),
            pyarrow.array(["d", "e", "f"]),
            pyarrow.array([False, True, False]),
        ],
        ["id", "name", "active"],
    )
    tables = [table1, table2]

    # Test the function with a limit of 4 rows
    rows, schema = converters.from_arrow(tables, 4)
    result = DataFrame(rows=rows, schema=schema)

    # Verify that the result has the correct number of rows and columns
    assert len(result) == 4
    assert len(result.column_names) == 3

    # Verify that the result has the correct column names and data types
    assert result.column_names == ("id", "name", "active")


def test_from_arrow_with_single_table():
    # Create a PyArrow table with some data
    data = [("Alice", 25), ("Bob", 30), ("Charlie", 35)]
    schema = pyarrow.schema([("name", pyarrow.string()), ("age", pyarrow.int64())])
    table = pyarrow.Table.from_pydict({f.name: [r[i] for r in data] for i, f in enumerate(schema)})

    # Create an instance of MyClass from the PyArrow table
    rows, schema = converters.from_arrow(table)
    obj = DataFrame(rows=rows, schema=schema)

    # Check that the instance has the correct rows and schema
    expected_rows = [("Alice", 25), ("Bob", 30), ("Charlie", 35)]
    expected_schema = {
        "name": {"type": str, "nullable": True},
        "age": {"type": int, "nullable": True},
    }
    obj.materialize()
    assert obj._rows == expected_rows, obj._rows
    assert obj._schema == expected_schema, obj._schema


def test_from_arrow_with_multiple_tables():
    # Create two PyArrow tables with some data
    data1 = [("Alice", 25), ("Bob", 30), ("Charlie", 35)]
    data2 = [("Dan", 40), ("Eve", 45), ("Frank", 50)]
    data3 = [("Ava", 15), ("Ethan", 20), ("Olivia", 55)]
    schema1 = pyarrow.schema([("name", pyarrow.string()), ("age", pyarrow.int64())])
    schema2 = pyarrow.schema([("name", pyarrow.string()), ("age", pyarrow.int64())])
    schema3 = pyarrow.schema([("name", pyarrow.string()), ("age", pyarrow.int64())])
    table1 = pyarrow.Table.from_pydict(
        {f.name: [r[i] for r in data1] for i, f in enumerate(schema1)}
    )
    table2 = pyarrow.Table.from_pydict(
        {f.name: [r[i] for r in data2] for i, f in enumerate(schema2)}
    )
    table3 = pyarrow.Table.from_pydict(
        {f.name: [r[i] for r in data3] for i, f in enumerate(schema3)}
    )

    # Create an instance of MyClass from a generator of PyArrow tables
    tables = (table1, table2, table3)
    rows, schema = converters.from_arrow(tables)
    obj = DataFrame(rows=rows, schema=schema)

    # Check that the instance has the correct rows and schema
    expected_rows = [
        ("Alice", 25),
        ("Bob", 30),
        ("Charlie", 35),
        ("Dan", 40),
        ("Eve", 45),
        ("Frank", 50),
        ("Ava", 15),
        ("Ethan", 20),
        ("Olivia", 55),
    ]
    expected_schema = {
        "name": {"type": str, "nullable": True},
        "age": {"type": int, "nullable": True},
    }
    obj.materialize()
    assert obj._rows == expected_rows, obj._rows
    assert obj._schema == expected_schema


def test_from_arrow_none():
    rows, schema = converters.from_arrow(None)
    assert rows == []
    assert schema == {}


def test_opteryx_arrow_small():
    import opteryx
    import orso

    planets = opteryx.query("SELECT * FROM $planets")
    planets_arrow = planets.arrow()
    assert isinstance(planets_arrow, pyarrow.Table)
    assert planets_arrow.shape == (9, 20)

    planets2 = orso.DataFrame.from_arrow(planets_arrow)
    assert isinstance(planets2, orso.DataFrame)
    assert planets2.shape == (9, 20)


def test_opteryx_arrow_medium():
    import opteryx
    import orso

    fake = opteryx.query("SELECT * FROM FAKE(100000, 100);")
    fake_arrow = fake.arrow()
    assert isinstance(fake_arrow, pyarrow.Table)
    assert fake_arrow.shape == (100000, 100), fake_arrow.shape

    fake2 = orso.DataFrame.from_arrow(fake_arrow)
    assert isinstance(fake2, orso.DataFrame)
    assert fake2.shape == (100000, 100)


if __name__ == "__main__":  # pragma: no cover
    test_from_arrow()
    test_from_arrow_with_single_table()
    test_from_arrow_with_multiple_tables()
    test_from_arrow_none()
    test_opteryx_arrow_small()
    test_opteryx_arrow_medium()
    print("✅ okay")



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso.tools import random_int
from orso.tools import random_string


def test_random():
    collected = [random_int() for i in range(1000000)]
    # allow some collisions - they're statistically likely here
    assert len(set(collected)) > (len(collected) * 0.999), len(set(collected))

    collected = [random_string() for i in range(1000000)]
    # allow one collisions, they're not impossible, but are unlikely
    assert len(set(collected)) >= len(collected) - 1
    assert all(len(c) == 16 for c in collected)


if __name__ == "__main__":  # pragma: no cover
    test_random()

    print("✅ okay")



import datetime
import functools

import numpy
import orjson

import orso

MAX_STRING_SIZE: int = 64
MAX_UNIQUE_COLLECTOR: int = 8


class DataProfile(orso.DataFrame):
    def __add__(self, data_profile):
        raise NotImplementedError("cannot add profiles")

    @classmethod
    def from_dataset(cls, dataset):
        _profile = table_profiler(dataset)
        profile = cls(_profile)
        profile._schema = {
            "name": {"type": str},
            "type": {"type": str},
            "count": {"type": int},
            "missing": {"type": int},
            "most_frequent_values": {"type": list},
            "most_frequent_counts": {"type": list},
            "numeric_range": {"type": list},
            "varchar_range": {"type": list},
            "distogram_values": {"type": list},
            "distogram_counts": {"type": list},
        }
        return profile


UNIX_EPOCH = datetime.datetime(1970, 1, 1)

TYPE_MAP = {
    "bool": "BOOLEAN",
    "bytes": "BLOB",
    "date": "DATE",
    "datetime": "TIMESTAMP",
    "dict": "STRUCT",
    "float": "DOUBLE",
    "int": "INTEGER",
    "list": "ARRAY",
    "str": "VARCHAR",
    "time": "TIME",
}


def _to_unix_epoch(date):
    if date is None:
        return None
    # Not all platforms can handle negative timestamp()
    # https://bugs.python.org/issue29097
    return (date - UNIX_EPOCH).total_seconds()


def table_profiler(dataframe):
    """
    Collect summary statistics about each column
    """
    from orso.profiler.distogram import Distogram  # type:ignore

    empty_profile = orjson.dumps(
        {
            "name": None,
            "type": [],
            "count": 0,
            "missing": 0,
            "most_frequent_values": None,
            "most_frequent_counts": None,
            "numeric_range": None,
            "varchar_range": None,
            "distogram_values": None,
            "distogram_counts": None,
        }
    )

    for morsel in dataframe.to_batches(10000):
        uncollected_columns = []
        profile_collector: dict = {}

        for column in morsel.column_names:
            column_data = morsel.collect(column)
            if len(column_data) == 0:
                continue

            profile = profile_collector.get(column, orjson.loads(empty_profile))
            _type: set = functools.reduce(
                lambda a, b: a if b is None else a.union({type(b).__name__}), column_data, set()
            )  # type:ignore

            profile["type"] = list(set(profile["type"]).union(_type))
            if len(profile["type"]) > 1:
                uncollected_columns.append(column)
            if len(_type) > 0:
                _type = _type.pop()
            else:
                _type = None
            profile["count"] += len(column_data)
            profile["missing"] += sum(1 for a in column_data if a is None)

            # interim save
            profile_collector[column] = profile

            # don't collect problematic columns
            if column in uncollected_columns:
                continue

            # don't collect columns we can't analyse
            if _type in {"list", "dict", "NoneType"}:
                continue

            # long strings are meaningless
            if _type == "str":
                column_data = [v for v in column_data if v is not None]

                max_len = functools.reduce(
                    lambda x, y: max(len(y), x),
                    column_data,
                    0,
                )
                if max_len > MAX_STRING_SIZE:
                    if column not in uncollected_columns:
                        uncollected_columns.append(column)
                    continue

                # collect the range values
                if len(column_data) > 0:
                    varchar_range_min = min(column_data)
                    varchar_range_max = max(column_data)

                    if profile["varchar_range"] is not None:
                        varchar_range_min = min(varchar_range_min, profile["varchar_range"][0])
                        varchar_range_max = max(varchar_range_max, profile["varchar_range"][1])

                    profile["varchar_range"] = (
                        varchar_range_min,
                        varchar_range_max,
                    )

            # convert TIMESTAMP into a NUMERIC (seconds after Unix Epoch)
            if _type == "datetime":
                column_data = (_to_unix_epoch(i) for i in column_data)

            if _type in {"bool", "datetime", "int", "float", "str"}:
                # remove empty values
                column_data = numpy.array([i for i in column_data if i not in (None, numpy.nan)])

            if _type == "bool":
                # we can make it easier to collect booleans
                counter = profile.get("counter")
                if counter is None:
                    counter = {"True": 0, "False": 0}
                trues = sum(column_data)
                counter["True"] += trues
                counter["False"] += column_data.size - trues
                profile["counter"] = counter

            if _type == "str" and profile.get("counter") != {}:
                # counter is used to collect and count unique values
                vals, counts = numpy.unique(column_data, return_counts=True)
                counter = {}
                if len(vals) <= MAX_UNIQUE_COLLECTOR:
                    counter = dict(zip(vals, counts))
                    for k, v in profile.get("counter", {}).items():
                        counter[k] = counter.pop(k, 0) + v
                    if len(counter) > MAX_UNIQUE_COLLECTOR:
                        counter = {}
                profile["counter"] = counter

            if _type in {"int", "float", "datetime"}:
                # populate the distogram, this is used for distribution statistics
                dgram = profile.get("dgram")
                if dgram is None:
                    dgram = Distogram()  # type:ignore
                dgram.bulkload(column_data)
                profile["dgram"] = dgram

            profile_collector[column] = profile

        buffer = []

        for column, profile in profile_collector.items():
            profile["name"] = column
            profile["type"] = ", ".join([TYPE_MAP.get(t, "OTHER") for t in profile["type"]])

            if column not in uncollected_columns:
                dgram = profile.pop("dgram", None)
                if dgram:
                    # force numeric types to be the same
                    profile["numeric_range"] = (
                        numpy.double(dgram.min),
                        numpy.double(dgram.max),
                    )
                    profile["distogram_values"], profile["distogram_counts"] = zip(*dgram.bins)
                    profile["distogram_values"] = list(
                        numpy.array(profile["distogram_values"], numpy.double)
                    )

                counter = profile.pop("counter", None)
                if counter:
                    counts = list(counter.values())
                    if min(counts) != max(counts):
                        profile["most_frequent_values"] = [str(k) for k in counter.keys()]
                        profile["most_frequent_counts"] = counts

            # remove collectors
            profile.pop("dgram", None)
            profile.pop("counter", None)

            buffer.append(profile)

        return buffer


if __name__ == "__main__":  # pragme: no cover
    # fmt:off
    cities_list:list = [
        {"name": "Tokyo", "population": 13929286, "country": "Japan", "founded": "1457", "area": 2191, "language": "Japanese"},
        {"name": "London", "population": 8982000, "country": "United Kingdom", "founded": "43 AD", "area": 1572, "language": "English"},
        {"name": "New York City", "population": 8399000, "country": "United States", "founded": "1624", "area": 468.9, "language": "English"},
        {"name": "Mumbai", "population": 18500000, "country": "India", "founded": "7th century BC", "area": 603.4, "language": "Hindi, English"},
        {"name": "Cape Town", "population": 433688, "country": "South Africa", "founded": "1652", "area": 400, "language": "Afrikaans, English"},
        {"name": "Paris", "population": 2148000, "country": "France", "founded": "3rd century BC", "area": 105.4, "language": "French"},
        {"name": "Beijing", "population": 21710000, "country": "China", "founded": "1045", "area": 16410.54, "language": "Mandarin"},
        {"name": "Rio de Janeiro", "population": 6747815, "country": "Brazil", "founded": "1 March 1565", "area": 1264, "language": "Portuguese"}
    ]
    import orso
    cities = orso.DataFrame(cities_list)
    # fmt:on

    import os
    import sys

    sys.path.insert(1, os.path.join(sys.path[0], "../.."))

    import opteryx

    df = opteryx.query("SELECT graduate_major, len(graduate_major) FROM $astronauts")
    print(df)

    from pprint import pprint

    pprint(table_profiler(df))



import os
import sys

sys.path.insert(1, os.path.join(sys.path[0], ".."))

from orso import cityhash


def test_cityhash_32():
    # Test data and expected hash values
    test_data = [
        (b"", 3696677242),
        (b"Hello, world!", 1877063093),
        (b"Lorem ipsum dolor sit amet, consectetur adipiscing elit.", 3701358904),
    ]

    for data, expected_hash in test_data:
        assert cityhash.CityHash32(data) == expected_hash, f"{data} - {cityhash.CityHash32(data)}"


def test_cityhash_64():
    # Test data and expected hash values
    test_data = [
        (b"", 11160318154034397263),
        (b"Hello, world!", 3493709964939663943),
        (b"Lorem ipsum dolor sit amet, consectetur adipiscing elit.", 11377970064775502085),
    ]

    for data, expected_hash in test_data:
        assert cityhash.CityHash64(data) == expected_hash


if __name__ == "__main__":  # pragma: no cover
    test_cityhash_32()
    test_cityhash_64()

    print("✅ okay")
